# Pipeline base class generation

This directory contains scripts and definitions to generate the `<TASK>Pipeline` base classes for the supported tasks.

## Supported tasks

All supported tasks are listed in `tasks.mk`.

## Generation process

`Pipeline` classes are generated from their [JSON schema](https://github.com/huggingface/huggingface.js/tree/main/packages/tasks/src/tasks) in two steps.

### Generating input and output dataclasses

First, the pipeline input and output normalized dataclasses are generated from the spec `input.json` and `output.json` into `input.py` and `output.py` files,
using [datamodel-code-generator](https://github.com/koxudaxi/datamodel-code-generator).

Each input and output dataclass inherits from `pydantic.BaseModel`, meaning that:

- each field is fully annotated,
- the dataclass constructor enforces type constraints.

### Generating pipeline classes

All `<TASK>Pipeline` base classes are defined using the same pattern.

```python
class MYTASKPipeline:
    def __call__(
        self,
        first_input: str,
        second_input: str,
        first_param: Optional[int] = 0,
        second_param: Optional[str] = 'default',
    ):
        normalized_inputs = MYTASKInputs(
            first_input=first_input,
            second_input=second_input,
        )
        normalized_parameters = MYTASKParameters(
            first_param=first_param,
            second_param=second_param,
        )
        return self.forward(normalized_inputs, parameters=normalized_parameters)

    def forward(self, inputs: MYTASKInputs, parameters: MYTASKParameters):
        raise NotImplementedError
```

They are generated by
- parsing the task input types to identify positional (mandatory) and keywords (optional) parameters,
- using them to define the `__call__` signature (include type hints and default values),
- using them to instantiate actual input dataclasses inside the `__call__` to enforce type constraints,
- using them to generate the signature of the `forward` method.

The backends implementing the actual `<Backend><Task>Pipeline` classes must inherit from these base class and only need to provide the `forward` implementation.

## Adding a new base class for a task

This should be as simple as adding the task to the list of supported tasks in `tasks.mk`.

However, if the task specification is too different from the other classes, the parsing code in `generate_pipeline.py` might need to be adapted.
